'use strict';
const {promisify} = require('util');
const path = require('path');
const fs = require('graceful-fs');
const writeFileAtomic = require('write-file-atomic');
const sortKeys = require('sort-keys');
const makeDir = require('make-dir');
const detectIndent = require('detect-indent');

const readFile = promisify(fs.readFile);
const writeFileAtomicP = promisify(writeFileAtomic);

const init = (fn, filePath, data, options) => {
	if (!filePath) {
		throw new TypeError('Expected a filepath');
	}

	if (data === undefined) {
		throw new TypeError('Expected data to stringify');
	}

	options = {
		indent: '\t',
		sortKeys: false,
		...options
	};

	if (options.sortKeys) {
		data = sortKeys(data, {
			deep: true,
			compare: typeof options.sortKeys === 'function' ? options.sortKeys : undefined
		});
	}

	return fn(filePath, data, options);
};

const main = async (filePath, data, options) => {
	let {indent} = options;

	if (options.detectIndent) {
		try {
			const file = await readFile(filePath, 'utf8');
			indent = detectIndent(file).indent;
		} catch (error) {
			if (error.code !== 'ENOENT') {
				throw error;
			}
		}
	}

	const json = JSON.stringify(data, options.replacer, indent);

	return writeFileAtomicP(filePath, `${json}\n`, {mode: options.mode});
};

const mainSync = (filePath, data, options) => {
	let {indent} = options;

	if (options.detectIndent) {
		try {
			const file = fs.readFileSync(filePath, 'utf8');
			indent = detectIndent(file).indent;
		} catch (error) {
			if (error.code !== 'ENOENT') {
				throw error;
			}
		}
	}

	const json = JSON.stringify(data, options.replacer, indent);

	return writeFileAtomic.sync(filePath, `${json}\n`, {mode: options.mode});
};

module.exports = async (filePath, data, options) => {
	await makeDir(path.dirname(filePath), {fs});
	return init(main, filePath, data, options);
};

module.exports.sync = (filePath, data, options) => {
	makeDir.sync(path.dirname(filePath), {fs});
	init(mainSync, filePath, data, options);
};
